from threading import Timer, Event, Lock


class DfxTimer(Timer):

    def __init__(self, interval, function, callback=None, args=None, kwargs=None):
        self.callback = callback
        super(DfxTimer, self).__init__(interval, function, args=args, kwargs=kwargs)

    def run(self):
        self.finished.wait(self.interval)
        if not self.finished.is_set():
            result = self.function(*self.args, **self.kwargs)
            self.callback(result)
        self.finished.set()


class DfxExecutor:

    def __init__(self, name, interval, function, callback=None, args=None, kwargs=None):
        self.name = name
        self.args = args
        self.kwargs = kwargs
        self.interval = interval
        self.function = function
        self.callback = callback
        self.finished = Event()

    def __start(self, _callback):
        DfxTimer(self.interval, self.function, _callback, self.args, self.kwargs).start()

    def run(self):
        def inline_callback(result):
            if self.callback:
                self.callback(result)
            DfxExecutorManager.cancel(self.name)

        if not self.finished.is_set():
            self.__start(inline_callback)

    def run_cycle(self):
        def inline_callback(result):
            if self.callback:
                self.callback(result)
            if not self.finished.is_set():
                self.__start(inline_callback)
            else:
                DfxExecutorManager.cancel(self.name)

        if not self.finished.is_set():
            self.__start(inline_callback)

    def cancel(self):
        self.finished.set()

    def is_finished(self):
        return self.finished.is_set()


class DfxExecutorManager:
    Executors = dict()
    ManagerLock = Lock()

    @classmethod
    def get_executor(cls, name, interval, function, callback=None, args=None, kwages=None):
        executor = DfxExecutor(name, interval, function, callback, args, kwages)
        cls.Executors[name] = executor
        return executor

    @classmethod
    def cancel(cls, name):
        with cls.ManagerLock:
            if name in cls.Executors:
                cls.Executors[name].cancel()
                del cls.Executors[name]

    @classmethod
    def check_state(cls, name):
        with cls.ManagerLock:
            if name in cls.Executors:
                return cls.Executors[name].is_finished()


import time, calendar
from datetime import datetime, timedelta


def today():
    """ 今天 """
    return datetime.today()


def today_date():
    """ 今天的日期 """
    return today().date()


def int_time_now():
    """ 时间戳 """
    return int(time.time())


def year(dt=None):
    """ 日期所在的年份 """
    if dt and isinstance(dt, datetime):
        return [dt.year]
    else:
        return year(today())


def year_month(dt=None):
    """ 日期所在的年月 """
    if dt and isinstance(dt, datetime):
        ret_lst = year(dt)
        ret_lst.append(dt.month)
        return ret_lst
    else:
        return year_month(today())


def year_month_day(dt=None):
    """ 日期所在的年月日 """
    if dt and isinstance(dt, datetime):
        ret_lst = year_month(dt)
        ret_lst.append(dt.day)
        return ret_lst
    else:
        return year_month_day(today())


def date_str(dt=None):
    """ 日期字符串 """
    if dt and isinstance(dt, datetime):
        return '-'.join([str(item) for item in year_month_day(dt)])
    else:
        return date_str(today())


def datetime_str(dt=None):
    """ 带时间的日期字符串 """
    if dt and isinstance(dt, datetime):
        return date_str(dt) + f' {dt.hour}:{dt.minute}:{dt.second}'
    else:
        return datetime_str(today())


def add_days(dt=None, days=None, minutes=None, seconds=None):
    """ 日期加减法 """
    if dt and isinstance(dt, datetime):
        if days and isinstance(days, int):
            dt += timedelta(days=days)
        if minutes and isinstance(minutes, int):
            dt += timedelta(minutes=minutes)
        if seconds and isinstance(seconds, int):
            dt += timedelta(seconds=seconds)
        return dt
    else:
        return add_days(today(), days, minutes, seconds)


def first_month_day(dt=None):
    """ 日期所在月的第一天的 """
    if dt and isinstance(dt, datetime):
        return dt.replace(day=1)
    else:
        return first_month_day(today())


def month_days(dt=None):
    """ 日期所在月有多少天 """
    if dt and isinstance(dt, datetime):
        return calendar.monthrange(*year_month(dt))[1]
    else:
        return month_days(today())



import random, string, hashlib, json, datetime
from typing import Any


def random_str(length: int) -> str:
    seed = string.digits + string.ascii_letters
    return ''.join(random.choices(seed, k=length))


def md5(_str: str):
    return hashlib.md5(_str.encode()).hexdigest()


def json_encoder(obj: Any):
    """ JSON 序列化, 修复时间 """
    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S')

    return super().default(obj)


def json_decoder(obj: Any):
    """ JSON 反序列化，加载时间 """
    ret = obj
    if isinstance(obj, list):
        obj = enumerate(obj)
    elif isinstance(obj, dict):
        obj = obj.items()
    else:
        return obj

    for key, item in obj:
        if isinstance(item, (list, dict)):
            ret[key] = json_decoder(item)
        elif isinstance(item, str):
            try:
                ret[key] = datetime.datetime.strptime(item, '%Y-%m-%d %H:%M:%S')
            except ValueError:
                ret[key] = item
        else:
            ret[key] = item
    return ret


def json_friendly_loads(obj: Any):
    return json.loads(obj, object_hook=json_decoder)


def json_friendly_dumps(obj: Any, **kwargs):
    return json.dumps(obj, ensure_ascii=False, default=json_encoder, **kwargs)


class DfxDict:

    def __init__(self, dict_ins=None):
        self.sep = '.'
        self.dict_ins = dict_ins or dict()

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        if self.sep in key:
            use_dict = self.dict_ins
            keys = key.split(self.sep)
            for item in keys:
                if item == keys[-1]:
                    use_dict[item] = value
                else:
                    use_dict[item] = {}
                    use_dict = use_dict[item]
        else:
            self.dict_ins[key] = value

    def __str__(self):
        return json_friendly_dumps(self.dict_ins)

    def get(self, key, default=None):
        if self.sep in key:
            ret_data = self.dict_ins
            keys = key.split(self.sep)
            for item in keys:
                if item in ret_data:
                    ret_data = ret_data.get(item, {})
                else:
                    return default
        else:
            ret_data = self.dict_ins.get(key, default)
        return DfxDict(ret_data) if isinstance(ret_data, dict) else ret_data


if __name__ == "__main__":
    tmp = DfxDict()
    tmp.sep = ','
    tmp['a,b.c.d'] = datetime.datetime.today()
    tmp = tmp['a']
    print(tmp)
